/*
 * Copyright (c) 2023-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifndef NETDEV_OFFLOAD_DPDK_H
#define NETDEV_OFFLOAD_DPDK_H

#include <sys/types.h>

#include "cmap.h"
#include "conntrack.h"
#include "dpdk-offload-provider.h"
#include "dpif.h"
#include "openvswitch/list.h"
#include "ovs-atomic.h"
#include "ovs-cancel.h"
#include "ovs-rcu.h"

struct unixctl_conn;
struct table_ctx;

struct act_resources {
    struct table_ctx *next_table_ctx;
    struct table_ctx *self_table_ctx;
    uint32_t flow_miss_ctx_id;
    uint32_t tnl_id;
    uint32_t flow_id;
    bool associated_flow_id;
    bool cont_in_sw;
    uint32_t ct_miss_ctx_id;
    uint32_t ct_match_zone_id;
    uint32_t ct_action_zone_id;
    uint32_t ct_match_label_id;
    uint32_t ct_action_label_id;
    uint32_t ct_action_label_ctx_id;
    struct indirect_ctx *shared_age_ctx;
    struct indirect_ctx *shared_count_ctx;
    uint32_t sflow_id;
    uint32_t meter_ids[DPDK_OFFLOAD_MAX_METERS_PER_FLOW];
};

struct ufid_to_rte_flow_data {
    union {
        struct cmap_node node;
        struct ovs_list list_node;
    };
    ovs_u128 ufid;
    struct netdev *netdev;
    struct flow_item flow_item;
    struct dpif_flow_stats stats;
    struct netdev *physdev;
    struct ovs_mutex lock;
    unsigned int creation_tid;
    struct ovsrcu_gc_node gc_node;
    atomic_bool active;
    struct act_resources act_resources;
};

static inline bool
rte_flow_data_active(struct ufid_to_rte_flow_data *data)
{
    bool active;

    if (!data) {
        return false;
    }
    atomic_read(&data->active, &active);
    return active;
}

static inline void
rte_flow_data_active_set(struct ufid_to_rte_flow_data *data, bool state)
{
    atomic_store(&data->active, state);
}

struct ct_offload_handle {
    struct ovs_cancel cancel;
    atomic_bool active[CT_DIR_NUM];
    struct ufid_to_rte_flow_data dir[CT_DIR_NUM];
};

/* Only CT_OFFLOAD_INIT and CT_OFFLOAD_REP flags are considered in 'dir'.
 * Return true if the flagged direction(s) offload is(are) currently active.
 * If CT_OFFLOAD_BOTH is set, then return true only if both are offloaded.
 */
static inline bool
ct_offload_handle_active(struct ct_offload_handle *coh, enum ct_offload_flag dir)
{
    bool ret = true;
    bool active;

    if (!coh) {
        return false;
    }

    if (dir & CT_OFFLOAD_INIT) {
        atomic_read(&coh->active[CT_DIR_INIT], &active);
        ret &= active;
    }

    if (dir & CT_OFFLOAD_REP) {
        atomic_read(&coh->active[CT_DIR_REP], &active);
        ret &= active;
    }

    return ret;
}

/* Only CT_OFFLOAD_INIT and CT_OFFLOAD_REP flags are considered in 'dir'.
 * Sets the respective mark for each set direction to 'state'.
 */
static inline void
ct_offload_handle_active_set(struct ct_offload_handle *coh, enum ct_offload_flag dir, bool state)
{
    if (dir & CT_OFFLOAD_INIT) {
        atomic_store(&coh->active[CT_DIR_INIT], state);
    }
    if (dir & CT_OFFLOAD_REP) {
        atomic_store(&coh->active[CT_DIR_REP], state);
    }
}

void
netdev_offload_dpdk_dump_traces(struct unixctl_conn *conn, int argc OVS_UNUSED,
                                const char *argv[] OVS_UNUSED, void *aux OVS_UNUSED);

int
netdev_offload_dpdk_get_conn_res(struct netdev *netdev,
                                 struct ct_flow_offload_item ct_offload[1],
                                 struct act_resources *act_resources);

void
netdev_offload_dpdk_free_conn_res(struct act_resources *act_resources);

#ifdef DPDK_NETDEV

#include <rte_flow.h>

/* Query the sw-meter installed on 'netdev' for the
 * number of dropped packets as a result of the rate-limiting.
 *
 * 'netdev': A netdev that has been initialized with the
 *           netdev-offload-dpdk offload provider.
 * 'query': The field 'hits' is set to the number of
 *          dropped packets, while 'bytes' is set to
 *          the number of dropped bytes.
 *
 * The return value is 0 on success, !0 otherwise.
 * If no sw-meter was configured on 'netdev', the value 0
 * is still returned and 'query' fields are set to 0.
 */
int
netdev_offload_dpdk_sw_meter_drops(const struct netdev *netdev,
                                   struct rte_flow_query_count *query);

#endif /* DPDK_NETDEV */

#endif /* NETDEV_OFFLOAD_DPDK_H */
