/*
 * Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef OVS_CANCEL_H
#define OVS_CANCEL_H

/* Asynchronous cancellation
 * -------------------------
 *
 *  This structure is used to render an asynchronous message
 *  cancellable by embedding it.
 *
 *  Usage:
 *
 *   struct foo {
 *      struct ovs_list node;
 *      struct ovs_cancel cancel;
 *   };
 *
 *   Thread 1                                  Thread 2
 *
 *   ovs_cancel_init(&foo->cancel);
 *   enqueue(foo);
 *   ...
 *   if (ovs_cancel(&foo->cancel)) {
 *       // Cancellation successful,
 *       // 'foo' not yet processed by T2.
 *   } else {
 *       // Cancelling too late,
 *       // 'foo' has already been processed.
 *   }
 *   ...
 *                                             foo = pop();
 *                                             if (!ovs_cancelled(&foo->cancel)) {
 *                                                 process(foo);
 *                                             }
 *                                             free(foo);
 *
 *  Only two threads can participate on this channel.
 */

#include "ovs-atomic.h"

struct ovs_cancel {
    struct ovs_refcount count;
};

static inline void
ovs_cancel_init(struct ovs_cancel *c)
{
    ovs_refcount_init(&c->count);
    ovs_refcount_ref(&c->count);
}

static inline bool
ovs_cancelled(struct ovs_cancel *c)
{
    return ovs_refcount_unref(&c->count) == 1;
}

static inline bool
ovs_cancel(struct ovs_cancel *c)
{
    return !ovs_cancelled(c);
}

#endif /* OVS_CANCEL_H */
