/*
 * Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software product is a proprietary product of NVIDIA CORPORATION &
 * AFFILIATES (the "Company") and all right, title, and interest in and to the
 * software product, including all associated intellectual property rights, are
 * and shall remain exclusively with the Company.
 *
 * This software product is governed by the End User License Agreement
 * provided with the software product.
 *
 */

#include <config.h>

#include <infiniband/verbs.h>
#include <unistd.h>

#include <rte_dev.h>
#include <rte_ethdev.h>
#include <rte_mtr.h>
#include <rte_pci.h>

#include <doca_dev.h>
#include <doca_dpdk.h>
#include <doca_flow.h>
#include <doca_log.h>
#include <doca_rdma_bridge.h>
#include <doca_version.h>

#include "coverage.h"
#include "conntrack-offload.h"
#include "dp-packet.h"
#include "dpdk.h"
#include "dpif-netdev.h"
#include "netdev-offload.h"
#include "netdev-offload-provider.h"
#include "openvswitch/dynamic-string.h"
#include "openvswitch/vlog.h"
#include "ovs-doca.h"
#include "refmap.h"
#include "unixctl.h"
#include "util.h"

VLOG_DEFINE_THIS_MODULE(ovs_doca);

COVERAGE_DEFINE(ovs_doca_no_mark);
COVERAGE_DEFINE(ovs_doca_invalid_classify_port);

#define OVS_DOCA_CONGESTION_THRESHOLD_DEFAULT 80
#define OVS_DOCA_CONGESTION_THRESHOLD_MIN 30
#define OVS_DOCA_CONGESTION_THRESHOLD_MAX 90
#define MAX_PORT_STR_LEN 128

/* Memorization counters are stored in pairs (red, green) in a contiguous array:
 *   ESW_0              ESW_1              ...
 *  [<------ N ------>][<------ N ------>][...]
 * Where N = OVS_DOCA_MAX_METER_COUNTERS_PER_ESW.
 */
static struct {
    struct doca_flow_query red;
    struct doca_flow_query green;
} mtr_last_stats[OVS_DOCA_MAX_METERS];

OVS_ASSERT_PACKED(struct ovs_doca_esw_key,
    struct rte_pci_addr rte_pci;
);

struct ovs_doca_esw_ctx_arg {
    struct ovs_doca_esw_key *esw_key;
    struct ovs_doca_netdev_data *dev_data;
};

/* Indicates successful initialization of DOCA. */
static atomic_bool doca_enabled = ATOMIC_VAR_INIT(false);
static atomic_bool doca_initialized = ATOMIC_VAR_INIT(false);
static FILE *log_stream = NULL;       /* Stream for DOCA log redirection */
static struct doca_log_backend *ovs_doca_log = NULL;
static struct refmap *ovs_doca_esw_rfm;
static struct ovs_mutex mgmt_queue_lock = OVS_MUTEX_INITIALIZER;

unsigned int doca_congestion_threshold = OVS_DOCA_CONGESTION_THRESHOLD_DEFAULT;
bool ovs_doca_async = true;
/* Size of control pipes. If zero, DOCA uses its default value. */
uint32_t ctl_pipe_size = 0;
uint32_t ctl_pipe_infra_size = 0;
atomic_bool ovs_doca_eswitch_active_ids[OVS_DOCA_MAX_ESW];

static void
ovs_doca_esw_port_uninit(struct ovs_doca_esw_ctx *ctx);

#define OVS_DOCA_MAX_SW_METER_COUNTERS RTE_MAX_ETHPORTS
#define OVS_DOCA_MAX_COUNTERS \
    (OVS_DOCA_MAX_MEGAFLOWS_COUNTERS + \
     OVS_DOCA_MAX_SW_METER_COUNTERS)

static inline unsigned int
ovs_doca_meter_counters_per_esw_base_id(void)
{
    return ovs_doca_max_ct_counters_per_esw();
}

bool
ovs_doca_enabled(void)
{
    bool enabled;

    atomic_read_relaxed(&doca_enabled, &enabled);
    return enabled;
}

bool
ovs_doca_initialized(void)
{
    bool initialized;

    atomic_read_relaxed(&doca_initialized, &initialized);
    return initialized;
}

static ssize_t
ovs_doca_log_write(void *c OVS_UNUSED, const char *buf, size_t size)
{
    static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(600, 600);
    static struct vlog_rate_limit dbg_rl = VLOG_RATE_LIMIT_INIT(600, 600);

    switch (doca_log_level_get_global_sdk_limit()) {
        case DOCA_LOG_LEVEL_TRACE:
            VLOG_DBG_RL(&dbg_rl, "%.*s", (int) size, buf);
            break;
        case DOCA_LOG_LEVEL_DEBUG:
            VLOG_DBG_RL(&dbg_rl, "%.*s", (int) size, buf);
            break;
        case DOCA_LOG_LEVEL_INFO:
            VLOG_INFO_RL(&rl, "%.*s", (int) size, buf);
            break;
        case DOCA_LOG_LEVEL_WARNING:
            VLOG_WARN_RL(&rl, "%.*s", (int) size, buf);
            break;
        case DOCA_LOG_LEVEL_ERROR:
            VLOG_ERR_RL(&rl, "%.*s", (int) size, buf);
            break;
        case DOCA_LOG_LEVEL_CRIT:
            VLOG_EMER("%.*s", (int) size, buf);
            break;
        default:
            OVS_NOT_REACHED();
    }

    return size;
}

static cookie_io_functions_t ovs_doca_log_func = {
    .write = ovs_doca_log_write,
};

static void
ovs_doca_unixctl_mem_stream(struct unixctl_conn *conn, int argc OVS_UNUSED,
                            const char *argv[] OVS_UNUSED, void *aux)
{
    void (*callback)(FILE *) = aux;
    char *response = NULL;
    FILE *stream;
    size_t size;

    stream = open_memstream(&response, &size);
    if (!stream) {
        response = xasprintf("Unable to open memstream: %s.",
                             ovs_strerror(errno));
        unixctl_command_reply_error(conn, response);
        goto out;
    }

    callback(stream);
    fclose(stream);
    unixctl_command_reply(conn, response);
out:
    free(response);
}

static const char * const levels[] = {
    [DOCA_LOG_LEVEL_CRIT]    = "critical",
    [DOCA_LOG_LEVEL_ERROR]   = "error",
    [DOCA_LOG_LEVEL_WARNING] = "warning",
    [DOCA_LOG_LEVEL_INFO]    = "info",
    [DOCA_LOG_LEVEL_DEBUG]   = "debug",
    [DOCA_LOG_LEVEL_TRACE]   = "trace",
};

static int
ovs_doca_parse_log_level(const char *s)
{
    int i;

    for (i = 0; i < ARRAY_SIZE(levels); ++i) {
        if (levels[i] && !strcmp(s, levels[i])) {
            return i;
        }
    }
    return -1;
}

static const char *
ovs_doca_log_level_to_str(uint32_t log_level)
{
    int i;

    for (i = 0; i < ARRAY_SIZE(levels); ++i) {
        if (i == log_level && levels[i]) {
            return levels[i];
        }
    }
    return NULL;
}

static void
ovs_doca_unixctl_log_set(struct unixctl_conn *conn, int argc,
                         const char *argv[], void *aux OVS_UNUSED)
{
    int level = DOCA_LOG_LEVEL_DEBUG;

    /* With no argument, level is set to 'debug'. */

    if (argc == 2) {
        const char *level_string;

        level_string = argv[1];
        level = ovs_doca_parse_log_level(level_string);
        if (level == -1) {
            char *err_msg;

            err_msg = xasprintf("invalid log level: '%s'", level_string);
            unixctl_command_reply_error(conn, err_msg);
            free(err_msg);
            return;
        }
    }

    doca_log_level_set_global_sdk_limit(level);
    unixctl_command_reply(conn, NULL);
}

static void
ovs_doca_log_dump(FILE *stream)
{
    uint32_t log_level;

    log_level = doca_log_level_get_global_sdk_limit();
    fprintf(stream, "DOCA log level is %s", ovs_doca_log_level_to_str(log_level));
}

static void
ovs_doca_dynamic_config(const struct smap *config)
{
    static const char *mode_names[] = {
        [0] = "synchronous",
        [1] = "asynchronous",
    };
    uint32_t req_ctl_pipe_size;
    bool req_doca_async;

    if (!smap_get_bool(config, "doca-init", false)) {
        return;
    }

    /* Once the user request a DOCA run, we must modify all future logic
     * (DPDK port probing) to take it into account, even if it results
     * in a failure. Once set, this value won't change. */
    if (!ovs_doca_enabled()) {
        atomic_store(&doca_enabled, true);
    }

    req_doca_async = smap_get_bool(config, "doca-async", true);
    if (req_doca_async != ovs_doca_async) {
        VLOG_INFO("Changing DOCA insertion mode from %s to %s.",
                  mode_names[!!ovs_doca_async], mode_names[!!req_doca_async]);

        ovs_doca_async = req_doca_async;
    } else {
        VLOG_INFO_ONCE("DOCA insertion mode is %s",
                       mode_names[!!ovs_doca_async]);
    }

    req_ctl_pipe_size = smap_get_uint(config, "ctl-pipe-size", 0);
    if (req_ctl_pipe_size != ctl_pipe_size) {
        VLOG_INFO("Changing DOCA ctl-pipe size from %"PRIu32" to %"PRIu32,
                  ctl_pipe_size, req_ctl_pipe_size);

        ctl_pipe_size = req_ctl_pipe_size;
    } else {
        VLOG_INFO_ONCE("DOCA ctl-pipe-size is %"PRIu32, ctl_pipe_size);
    }
}

static void
ovs_doca_pipe_uninit(struct doca_flow_pipe **ppipe)
{
    if (!*ppipe) {
        return;
    }

    doca_flow_pipe_destroy(*ppipe);
    *ppipe = NULL;
}

static void
ovs_doca_entry_uninit(struct doca_flow_pipe_entry **pentry)
{
    if (!*pentry) {
        return;
    }

    ovs_doca_mgmt_queue_lock();
    doca_flow_pipe_rm_entry(AUX_QUEUE, DOCA_FLOW_NO_WAIT, *pentry);
    ovs_doca_mgmt_queue_unlock();
    *pentry = NULL;
}

static struct doca_flow_pipe *
ovs_doca_egress_pipe_init(struct ovs_doca_netdev_data *dev_data)
{
    const char *pci = dev_data->esw_ctx->pci_addr;
    struct doca_flow_pipe *pipe = NULL;
    struct doca_flow_pipe_cfg *cfg;
    struct doca_flow_match match;
    struct doca_flow_fwd fwd;
    char pipe_name[50];
    int ret;

    memset(&match, 0, sizeof(match));
    memset(&fwd, 0, sizeof(fwd));

    /* Meta to match on is defined per entry */
    match.meta.pkt_meta = UINT32_MAX;
    /* Port ID to forward to is defined per entry */
    fwd.type = DOCA_FLOW_FWD_PORT;
    fwd.port_id = UINT16_MAX;
    snprintf(pipe_name, sizeof pipe_name, "OVS_EGRESS_PIPE_%s", pci);

    ret = doca_flow_pipe_cfg_create(&cfg, dev_data->port);
    if (ret) {
        VLOG_ERR("%s: Could not create doca_flow_pipe_cfg for egress", pci);
        return NULL;
    }

    if (doca_flow_pipe_cfg_set_name(cfg, pipe_name) ||
        doca_flow_pipe_cfg_set_type(cfg, DOCA_FLOW_PIPE_BASIC) ||
        doca_flow_pipe_cfg_set_domain(cfg, DOCA_FLOW_PIPE_DOMAIN_EGRESS) ||
        doca_flow_pipe_cfg_set_match(cfg, &match, &match) ||
        doca_flow_pipe_cfg_set_is_root(cfg, true) ||
        doca_flow_pipe_cfg_set_nr_entries(cfg, RTE_MAX_ETHPORTS)) {
        VLOG_ERR("%s: Could not set doca_flow_pipe_cfg for egress", pci);
        goto out;
    }

    ret = doca_flow_pipe_create(cfg, &fwd, NULL, &pipe);
    if (ret) {
        VLOG_ERR("%s: Failed to create egress pipe. Error %d (%s)", pci, ret,
                 doca_error_get_descr(ret));
        goto out;
    }

out:
    doca_flow_pipe_cfg_destroy(cfg);
    return pipe;
}

static void
ovs_doca_rss_uninit(struct ovs_doca_esw_ctx *esw_ctx)
{
    ovs_doca_entry_uninit(&esw_ctx->rss_entry);
    ovs_doca_pipe_uninit(&esw_ctx->rss_pipe);
}

static int
ovs_doca_rss_init(struct ovs_doca_netdev_data *dev_data,
                  int num_of_queues)
{
    const char *pci = dev_data->esw_ctx->pci_addr;
    struct doca_flow_pipe_entry *entry;
    struct doca_flow_pipe *pipe = NULL;
    struct doca_flow_pipe_cfg *cfg;
    struct doca_flow_match match;
    struct doca_flow_fwd fwd;
    uint16_t *rss_queues;
    char pipe_name[50];
    int ret;
    int q;

    memset(&match, 0, sizeof(match));
    memset(&fwd, 0, sizeof(fwd));

    rss_queues = xcalloc(num_of_queues, sizeof *rss_queues);
    for (q = 0; q < num_of_queues; q++) {
        rss_queues[q] = q;
    }

    fwd.type = DOCA_FLOW_FWD_RSS;
    fwd.rss_queues = rss_queues;
    fwd.num_of_queues = num_of_queues;
    snprintf(pipe_name, sizeof pipe_name, "OVS_RSS_PIPE_%s", pci);

    ret = doca_flow_pipe_cfg_create(&cfg, dev_data->port);
    if (ret) {
        VLOG_ERR("%s: Could not create doca_flow_pipe_cfg for RSS", pci);
        goto out;
    }

    if (doca_flow_pipe_cfg_set_name(cfg, pipe_name) ||
        doca_flow_pipe_cfg_set_type(cfg, DOCA_FLOW_PIPE_BASIC) ||
        doca_flow_pipe_cfg_set_match(cfg, &match, &match) ||
        doca_flow_pipe_cfg_set_nr_entries(cfg, 1)) {
        VLOG_ERR("%s: Could not set doca_flow_pipe_cfg for RSS", pci);
        goto out;
    }

    ret = doca_flow_pipe_create(cfg, &fwd, NULL, &pipe);
    if (ret) {
        VLOG_ERR("%s: Failed to create rss pipe. Error %d (%s)", pci, ret,
                 doca_error_get_descr(ret));
        goto out;
    }

    ovs_doca_mgmt_queue_lock();
    ret = doca_flow_pipe_add_entry(AUX_QUEUE, pipe, &match, NULL, NULL,
                                   &fwd, DOCA_FLOW_NO_WAIT, NULL, &entry);
    ovs_doca_mgmt_queue_unlock();
    if (ret) {
        VLOG_ERR("%s: Failed to create rss entry: Error %d (%s)", pci, ret,
                 doca_error_get_descr(ret));
        goto out;
    }

    dev_data->esw_ctx->rss_pipe = pipe;
    dev_data->esw_ctx->rss_entry = entry;

out:
    doca_flow_pipe_cfg_destroy(cfg);
    if (ret) {
        ovs_doca_pipe_uninit(&pipe);
    }
    free(rss_queues);

    return ret;
}

static struct ds *
dump_ovs_doca_esw(struct ds *s, void *key_,
                  void *ctx_ OVS_UNUSED, void *arg_ OVS_UNUSED)
{
    struct ovs_doca_esw_key *key = key_;
    char pci_addr[PCI_PRI_STR_SIZE];

    rte_pci_device_name(&key->rte_pci, pci_addr, sizeof pci_addr);
    ds_put_format(s, "pci=%s, ", pci_addr);

    return s;
}

static int
ovs_doca_dev_open_pci(struct rte_pci_addr *rte_pci, struct doca_dev **pdev)
{
    struct doca_devinfo **dev_list;
    char pci[PCI_PRI_STR_SIZE];
    uint8_t is_addr_equal = 0;
    uint32_t nb_devs;
    size_t i;
    int res;

    /* Set default return value */
    *pdev = NULL;

    res = doca_devinfo_create_list(&dev_list, &nb_devs);
    if (res != DOCA_SUCCESS) {
        VLOG_ERR("Failed to load doca devices list. Doca_error value: %d",
                 res);
        return res;
    }

    rte_pci_device_name(rte_pci, pci, sizeof pci);
    /* Search */
    for (i = 0; i < nb_devs; i++) {
        res = doca_devinfo_is_equal_pci_addr(dev_list[i], pci, &is_addr_equal);
        if (res != DOCA_SUCCESS || !is_addr_equal) {
            continue;
        }

        VLOG_INFO("Opening '%s'", pci);
        res = doca_dev_open(dev_list[i], pdev);
        if (res != DOCA_SUCCESS) {
            VLOG_ERR("Failed to open DOCA device: %s",
                     doca_error_get_descr(res));
        }
        goto out;
    }

    VLOG_WARN("No matching doca device found");
    res = DOCA_ERROR_NOT_FOUND;

out:
    doca_devinfo_destroy_list(dev_list);
    return res;
}

static void
ovs_doca_esw_ctx_uninit(void *ctx_)
{
    struct ovs_doca_esw_ctx *ctx = ctx_;

    memset(ctx->pci_addr, 0, sizeof ctx->pci_addr);
}

static int
ovs_doca_esw_ctx_init(void *ctx_, void *arg_)
{
    struct ovs_doca_esw_ctx_arg *arg = arg_;
    struct ovs_doca_esw_ctx *ctx = ctx_;

    if (ovs_doca_dev_open_pci(&arg->esw_key->rte_pci, &ctx->dev)) {
        return ENODEV;
    }
    rte_pci_device_name(&arg->esw_key->rte_pci, ctx->pci_addr,
                        sizeof ctx->pci_addr);
    ctx->cmd_fd = -1;

    return 0;
}

int
ovs_doca_init(const struct smap *ovs_other_config)
{
    static struct ovsthread_once once_enable = OVSTHREAD_ONCE_INITIALIZER;
    unsigned int nb_threads = DEFAULT_OFFLOAD_THREAD_NB;
    unsigned int req_doca_congestion_threshold;
    struct doca_flow_resource_rss_cfg rss_cfg;
    uint32_t req_ctl_pipe_infra_size;
    static bool enabled = false;
    struct doca_flow_cfg *cfg;
    uint16_t rss_queues;
    doca_error_t err;

    /* Dynamic configuration:
     * This section can be modified without restarting the process. */

    ovs_doca_dynamic_config(ovs_other_config);

    /* Static configuration:
     * This section is set once, restart is required after a change. */

    if (!ovsthread_once_start(&once_enable)) {
        return 0;
    }

    req_ctl_pipe_infra_size = smap_get_uint(ovs_other_config,
                                            "ctl-pipe-infra-size", UINT32_MAX);
    if (req_ctl_pipe_infra_size != UINT32_MAX) {
        VLOG_INFO("Changing DOCA ctl-pipe infra size from %"PRIu32" to %"PRIu32,
                  ctl_pipe_infra_size, req_ctl_pipe_infra_size);

        ctl_pipe_infra_size = req_ctl_pipe_infra_size;
    } else {
        ctl_pipe_infra_size = ctl_pipe_size;
    }
    VLOG_INFO("DOCA ctl-pipe-infra_size is %"PRIu32, ctl_pipe_infra_size);

    log_stream = fopencookie(NULL, "w+", ovs_doca_log_func);
    if (log_stream == NULL) {
        VLOG_ERR("Can't redirect DOCA log: %s.", ovs_strerror(errno));
    } else {
        /* Create a logger backend that prints to the redirected log */
        err = doca_log_backend_create_with_file_sdk(log_stream, &ovs_doca_log);
        if (err != DOCA_SUCCESS) {
            ovsthread_once_done(&once_enable);
            return EXIT_FAILURE;
        }
        doca_log_level_set_global_sdk_limit(DOCA_LOG_LEVEL_WARNING);
    }
    unixctl_command_register("doca/log-set", "{level}. level=critical/error/"
                             "warning/info/debug", 0, 1,
                             ovs_doca_unixctl_log_set, NULL);
    unixctl_command_register("doca/log-get", "", 0, 0,
                             ovs_doca_unixctl_mem_stream, ovs_doca_log_dump);

    if (!enabled && ovs_other_config &&
        smap_get_bool(ovs_other_config, "doca-init", false)) {
        ovs_doca_esw_rfm = refmap_create("ovs-doca-esw",
                                         sizeof(struct ovs_doca_esw_key),
                                         sizeof(struct ovs_doca_esw_ctx),
                                         ovs_doca_esw_ctx_init,
                                         ovs_doca_esw_ctx_uninit,
                                         dump_ovs_doca_esw, false);

        /* Set dpdk-init to be true if not already set */
        smap_replace(CONST_CAST(struct smap *, ovs_other_config),
                     "dpdk-init", "true");
        dpdk_init(ovs_other_config);

        /* OVS-DOCA configuration happens earlier than dpif-netdev's.
         * To avoid reorganizing them, read the relevant item directly. */
        conntrack_offload_config(ovs_other_config);

        /* Due to limitation in doca, only one offload thread is currently
         * supported.
         * */
        if (smap_get_uint(ovs_other_config,"n-offload-threads", 1) !=
            nb_threads) {
            smap_replace(CONST_CAST(struct smap *, ovs_other_config),
                         "n-offload-threads", "1");
            VLOG_WARN_ONCE("Only %u offload thread is currently supported with"
                           "doca, ignoring n-offload-threads configuration",
                           nb_threads);
        }

        req_doca_congestion_threshold =
            smap_get_uint(ovs_other_config, "doca-congestion-threshold",
                          OVS_DOCA_CONGESTION_THRESHOLD_DEFAULT);
        if (req_doca_congestion_threshold != doca_congestion_threshold) {
            if (req_doca_congestion_threshold < OVS_DOCA_CONGESTION_THRESHOLD_MIN ||
                req_doca_congestion_threshold > OVS_DOCA_CONGESTION_THRESHOLD_MAX ) {
                VLOG_WARN("doca-congestion-threshold (%u) is not within expected"
                           " range (%u, %u)", req_doca_congestion_threshold,
                           OVS_DOCA_CONGESTION_THRESHOLD_MIN,
                           OVS_DOCA_CONGESTION_THRESHOLD_MAX);
            }
            VLOG_INFO("Changing DOCA doca-congestion-threshold from %u to %u",
                      doca_congestion_threshold, req_doca_congestion_threshold);

            doca_congestion_threshold = req_doca_congestion_threshold;
        } else {
            VLOG_INFO_ONCE("DOCA doca-congestion-threhold is %u",
                           doca_congestion_threshold);
        }

        ovs_assert(!doca_flow_cfg_create(&cfg));
        ovs_assert(!
            doca_flow_cfg_set_pipe_queues(cfg, smap_get_uint(ovs_other_config,
                                          "n-offload-threads", 1) + 1));
        ovs_assert(!
            doca_flow_cfg_set_nr_counters(cfg, OVS_DOCA_MAX_COUNTERS));
        ovs_assert(!doca_flow_cfg_set_mode_args(cfg, "switch,hws,cpds,isolated"));
        ovs_assert(!doca_flow_cfg_set_queue_depth(cfg, OVS_DOCA_QUEUE_DEPTH));
        ovs_assert(!doca_flow_cfg_set_cb_entry_process(cfg,
                                                       ovs_doca_entry_process_cb));
        ovs_assert(!doca_flow_cfg_set_cb_pipe_process(cfg,
                                                      ovs_doca_pipe_process_cb));
        memset(&rss_cfg, 0, sizeof rss_cfg);
        rss_cfg.nr_queues = 1;
        rss_queues = 0;
        rss_cfg.queues_array = &rss_queues;
        ovs_assert(!doca_flow_cfg_set_default_rss(cfg, &rss_cfg));
        /* Set the sum of counters we want for both ports */
        ovs_assert(!
            doca_flow_cfg_set_nr_shared_resource(cfg,
                                                 ovs_doca_max_shared_counters(),
                                                 DOCA_FLOW_SHARED_RESOURCE_COUNTER));
        ovs_assert(!
            doca_flow_cfg_set_nr_shared_resource(cfg,
                                                 OVS_DOCA_MAX_METERS,
                                                 DOCA_FLOW_SHARED_RESOURCE_METER));

        VLOG_INFO("DOCA Enabled - initializing...");
        err = doca_flow_init(cfg);
        ovs_assert(!doca_flow_cfg_destroy(cfg));
        if (err) {
            VLOG_ERR("Error initializing doca flow offload. Error %d (%s)\n",
                    err, doca_error_get_descr(err));

            ovsthread_once_done(&once_enable);
            ovs_abort(err, "Cannot init DOCA");
            return err;
        }

        enabled = true;
        VLOG_INFO("DOCA Enabled - initialized");
    }

    ovsthread_once_done(&once_enable);
    atomic_store_relaxed(&doca_initialized, enabled);

    return 0;
}

/* Extract the PCI part to the provided esw_key argument, and rte_devargs.
 * Return -EINVAL for error or 0 for success.
 */
static int
ovs_doca_esw_key_parse(const char *devargs,
                       struct ovs_doca_esw_key *esw_key)
{
    memset(esw_key, 0, sizeof *esw_key);

    if (!dpdk_parse_devargs_pci(devargs, &esw_key->rte_pci)) {
        return -1;
    }

    return 0;
}

static struct doca_flow_pipe_entry *
ovs_doca_egress_entry_init(struct doca_flow_port *port,
                           struct doca_flow_pipe *pipe,
                           int port_id)
{
    static struct doca_flow_pipe_entry *entry;
    struct doca_flow_match match;
    struct doca_flow_fwd fwd;
    doca_error_t err;

    memset(&match, 0, sizeof(match));
    memset(&fwd, 0, sizeof(fwd));

    match.meta.pkt_meta = ovs_doca_port_id_pkt_meta(port_id);

    fwd.type = DOCA_FLOW_FWD_PORT;
    fwd.port_id = port_id;

    ovs_doca_mgmt_queue_lock();

    err = doca_flow_pipe_add_entry(AUX_QUEUE, pipe, &match, NULL, NULL, &fwd,
                                   DOCA_FLOW_NO_WAIT, NULL,
                                   &entry);
    if (err) {
        VLOG_ERR("Failed to create egress pipe entry. Error %d (%s)", err,
                 doca_error_get_descr(err));
        entry = NULL;
        goto unlock;
    }

    err = doca_flow_entries_process(port, AUX_QUEUE, ENTRY_PROCESS_TIMEOUT_MS, 0);
    if (err) {
        VLOG_ERR("Failed to process egress pipe entry. Error: %d (%s)", err,
                 doca_error_get_descr(err));
        entry = NULL;
        goto unlock;
    }

unlock:
    ovs_doca_mgmt_queue_unlock();

    return entry;
}

static void
ovs_doca_esw_port_uninit(struct ovs_doca_esw_ctx *ctx)
{
    uint16_t pid;

    if (!ctx) {
        return;
    }

    for (pid = 0; pid < RTE_MAX_ETHPORTS; pid++) {
        if (ctx->port_queue_batch[pid]) {
            free(ctx->port_queue_batch[pid]);
            ctx->port_queue_batch[pid] = NULL;
        }
    }
    ovs_doca_rss_uninit(ctx);
    ovs_doca_pipe_uninit(&ctx->egress_pipe);
}

static int
ovs_doca_esw_port_init(struct ovs_doca_netdev_data *dev_data)
{
    struct ovs_doca_esw_ctx *ctx = dev_data->esw_ctx;
    unsigned int n_pmd_threads;
    uint16_t pid;

    n_pmd_threads = dpif_netdev_get_n_pmd_threads();
    if (n_pmd_threads != 1) {
        VLOG_WARN("%s: Multiple PMD threads configured (%d), which is not supported by DOCA. "
                  "DOCA ports will run on a single thread", ctx->pci_addr, n_pmd_threads);
        n_pmd_threads = 1;
    }

    ctx->egress_pipe = ovs_doca_egress_pipe_init(dev_data);
    if (ctx->egress_pipe == NULL) {
        goto err;
    }
    if (ovs_doca_rss_init(dev_data, n_pmd_threads)) {
        goto err;
    }

    ctx->port_id = dev_data->port_id;
    for (pid = 0; pid < RTE_MAX_ETHPORTS; pid++) {
        uint16_t qid;

        ctx->port_queue_batch[pid] = xcalloc(n_pmd_threads,
                                             sizeof(struct dp_packet_batch));
        if (!ctx->port_queue_batch[pid]) {
            goto err;
        }
        for (qid = 0; qid < n_pmd_threads; qid++) {
            dp_packet_batch_init(&ctx->port_queue_batch[pid][qid]);
        }
    }

    return 0;
err:
    ovs_doca_esw_port_uninit(ctx);
    return -1;
}

int
ovs_doca_port_start(struct ovs_doca_netdev_data *dev_data)
{
    const char *devargs = dev_data->devargs;
    uint16_t port_id = dev_data->port_id;
    struct doca_flow_port_cfg *port_cfg;
    char port_id_str[MAX_PORT_STR_LEN];
    struct ovs_doca_esw_ctx *esw_ctx;
    struct doca_flow_port *esw_port;
    int err;

    if (!ovs_doca_initialized()) {
        VLOG_ERR("DOCA is not initialized: Failed to initialize port %u",
                 port_id);
        return -1;
    }

    err = doca_flow_port_cfg_create(&port_cfg);
    if (err) {
        VLOG_ERR("Failed to create doca flow port_cfg. Error: %d (%s)",
                 err, doca_error_get_descr(err));
        return -1;
    }

    if (ovs_doca_dev_probe(dev_data, devargs, true)) {
        err = -1;
        goto out;
    }
    esw_ctx = dev_data->esw_ctx;
    if (!esw_ctx) {
        err = -1;
        goto out;
    }

    snprintf(port_id_str, MAX_PORT_STR_LEN, "%"PRIu16, port_id);
    err = doca_flow_port_cfg_set_devargs(port_cfg, port_id_str);
    if (err) {
        VLOG_ERR("Failed to set doca flow port_cfg devargs %s. Error: %d (%s)",
                 port_id_str, err, doca_error_get_descr(err));
        goto out;
    }
    err = doca_flow_port_cfg_set_dev(port_cfg, esw_ctx->dev);
    if (err) {
        VLOG_ERR("Failed to set doca flow port_cfg dev. Error: %d (%s)", err,
                 doca_error_get_descr(err));
        goto out;
    }

    VLOG_INFO("Starting '%s', port_id=%d", devargs, port_id);
    if (dev_data->port_id == dev_data->esw_mgr_port_id) {
        err = rte_eth_dev_start(dev_data->port_id);
        if (err) {
            VLOG_ERR("Failed to start dpdk port_id %"PRIu16". Error: %d (%s)",
                     dev_data->port_id, err, doca_error_get_descr(err));
            goto out;
        }
    }
    err = doca_flow_port_start(port_cfg, &dev_data->port);
    if (err) {
        VLOG_ERR("Failed to start doca flow port_id %"PRIu16". Error: %d (%s)",
                 port_id, err, doca_error_get_descr(err));
        goto out;
    }

    if (dev_data->port_id == dev_data->esw_mgr_port_id &&
        ovs_doca_esw_port_init(dev_data)) {
        err = -1;
        goto out;
    }
    esw_port = doca_flow_port_switch_get(dev_data->port);
    dev_data->egress_entry = ovs_doca_egress_entry_init(esw_port,
                                                        esw_ctx->egress_pipe,
                                                        port_id);
    if (!dev_data->egress_entry) {
        err = -1;
        goto out;
    }

out:
    doca_flow_port_cfg_destroy(port_cfg);
    if (err) {
        ovs_doca_port_stop(dev_data);
    }
    return err;
}

int
ovs_doca_port_stop(struct ovs_doca_netdev_data *dev_data)
{
    int err = 0;

    ovs_assert(ovs_doca_initialized());

    VLOG_INFO("Stopping '%s', port_id=%d", dev_data->devargs, dev_data->port_id);

    ovs_doca_entry_uninit(&dev_data->egress_entry);

    if (dev_data->port_id == dev_data->esw_mgr_port_id) {
        ovs_doca_esw_port_uninit(dev_data->esw_ctx);
    }

    if (dev_data->port) {
        err = doca_flow_port_stop(dev_data->port);
        dev_data->port = NULL;
        rte_eth_dev_stop(dev_data->port_id);
    }

    return err;
}

void
ovs_doca_set_properties(struct ovs_doca_netdev_data *dev_data,
                        uint16_t port_id,
                        const char *devargs)
{
    uint16_t proxy_port_id;

    dev_data->port_id = port_id;
    dev_data->esw_mgr_port_id = port_id;
    dev_data->devargs = devargs;
    /* For doca_netdev_data, get the proxy port from dpdk, as it might not
     * been probed yet in OVS.
     */
    if (rte_eth_dev_is_valid_port(port_id)) {
        rte_flow_pick_transfer_proxy(port_id, &proxy_port_id, NULL);
        dev_data->esw_mgr_port_id = proxy_port_id;
    }
}

void
ovs_doca_status(const struct ovsrec_open_vswitch *cfg)
{
    if (!cfg) {
        return;
    }
    ovsrec_open_vswitch_set_doca_initialized(cfg, ovs_doca_initialized());
    ovsrec_open_vswitch_set_doca_version(cfg, doca_version_runtime());
}

void
print_doca_version(void)
{
    printf("DOCA %s\n", doca_version_runtime());
}

const char *
ovs_doca_get_version(void)
{
    return doca_version_runtime();
}

static void
fill_meter_profile(struct doca_flow_shared_resource_cfg *doca_cfg,
                   struct ofputil_meter_config *config)
{
    doca_cfg->domain = DOCA_FLOW_PIPE_DOMAIN_DEFAULT;

    if (config->flags & OFPMF13_PKTPS) {
        doca_cfg->meter_cfg.limit_type = DOCA_FLOW_METER_LIMIT_TYPE_PACKETS;
        doca_cfg->meter_cfg.cir = config->bands[0].rate;
        doca_cfg->meter_cfg.cbs = config->bands[0].burst_size;
    } else {
        doca_cfg->meter_cfg.limit_type = DOCA_FLOW_METER_LIMIT_TYPE_BYTES;
        /* Convert from kilobits per second to bytes per second */
        doca_cfg->meter_cfg.cir = ((uint64_t) config->bands[0].rate) * 125;
        doca_cfg->meter_cfg.cbs = ((uint64_t) config->bands[0].burst_size) * 125;
    }
}

uint32_t
ovs_doca_meter_id(uint32_t dp_meter_id, uint32_t esw_id)
{
    return esw_id * OVS_DOCA_MAX_METERS_PER_ESW + dp_meter_id;
}

int
ovs_doca_create_meter(uint32_t of_meter_id,
                      struct ofputil_meter_config *config,
                      struct rte_mtr_error *error)
{
    struct doca_flow_shared_resource_cfg meter_cfg;
    uint32_t doca_mtr_id;
    uint32_t esw_id;
    int ret;

    if (config->n_bands != 1) {
        return -1;
    }

    memset(&meter_cfg, 0, sizeof meter_cfg);
    fill_meter_profile(&meter_cfg, config);

    /* At this stage it's not known with which eswitch the meter will be used,
     * thus need to configure meter for all available eswitches.
     */
    for (esw_id = 0; esw_id < OVS_DOCA_MAX_ESW; esw_id++) {
        /* id is determine by both the upper layer id, and the esw_id. */
        doca_mtr_id = ovs_doca_meter_id(of_meter_id, esw_id);

        ret = doca_flow_shared_resource_cfg(DOCA_FLOW_SHARED_RESOURCE_METER,
                                            doca_mtr_id, &meter_cfg);
        if (ret != DOCA_SUCCESS) {
            VLOG_ERR("Failed to configure shared meter id %d for eswitch%d, err %d - %s",
                     of_meter_id, esw_id, ret, doca_error_get_descr(ret));
            if (error) {
                error->type = RTE_MTR_ERROR_TYPE_UNSPECIFIED;
                error->message = doca_error_get_descr(ret);
                break;
            }
        }
    }

    return ret;
}

static int
color_offset(enum doca_flow_meter_color meter_color)
{
    if (meter_color == DOCA_FLOW_METER_COLOR_RED) {
        return 0;
    }
    if (meter_color == DOCA_FLOW_METER_COLOR_GREEN) {
        return 1;
    }
    /* Yellow color is not used at the moment */
    OVS_NOT_REACHED();
    return -1;
}

uint32_t
ovs_doca_get_post_meter_counter_id(uint32_t doca_meter_id,
                                   enum doca_flow_meter_color meter_color)
{
    uint32_t eswitch_id = doca_meter_id / OVS_DOCA_MAX_METERS_PER_ESW;
    int offset = color_offset(meter_color);

    return eswitch_id * ovs_doca_max_shared_counters_per_esw() +
           ovs_doca_meter_counters_per_esw_base_id() +
           (doca_meter_id - (eswitch_id * OVS_DOCA_MAX_METERS_PER_ESW)) *
           OVS_DOCA_MAX_METER_COLORS - offset;
}

static int
query_counter(uint32_t counter_id, struct doca_flow_query *counter_stats,
              struct rte_mtr_error *error)
{
    struct doca_flow_shared_resource_result query_results;
    doca_error_t ret;

    memset(&query_results, 0, sizeof query_results);

    ret = doca_flow_shared_resources_query(DOCA_FLOW_SHARED_RESOURCE_COUNTER,
                                           &counter_id, &query_results, 1);
    if (ret != DOCA_SUCCESS) {
        VLOG_ERR("Failed to query shared meter counter id %u: %s",
                 counter_id, doca_error_get_descr(ret));
        error->type = RTE_MTR_ERROR_TYPE_STATS;
        error->message = doca_error_get_descr(ret);
        return -1;
    }

    memcpy(counter_stats, &query_results.counter, sizeof *counter_stats);

    return 0;
}

int
ovs_doca_delete_meter(uint16_t port_id OVS_UNUSED, uint32_t dp_meter_id,
                      struct rte_mtr_error *error)
{
    struct doca_flow_query red_stats, green_stats;
    uint32_t red_counter_id, green_counter_id;
    uint32_t doca_mtr_id;
    uint32_t esw_id;
    int ret;

    /* DOCA shared meter cannot be unbound explicitly, instead DOCA handles it
    * internally, when the port or the pipe, where the meter has been bound to,
    * is destroyed.
    * Same holds true for meter counters. This means that the next time user
    * creates a meter with the same ID that has been used previously, the
    * statistics will be reported incorrectly as the meter counters cannot be
    * reset while they're bound.
    * As a workaround counter values are saved when meter is deleted and used
    * to offset statistics for the next meter that re-uses the same ID.
    */
    OVS_DOCA_FOREACH_ACTIVE_ESWITCH (esw_id) {
        /* id is determine by both the upper layer id, and the esw_id. */
        doca_mtr_id = ovs_doca_meter_id(dp_meter_id, esw_id);

        red_counter_id = ovs_doca_get_post_meter_counter_id(doca_mtr_id,
                                                            DOCA_FLOW_METER_COLOR_RED);
        ret = query_counter(red_counter_id, &red_stats, error);
        if (ret) {
            VLOG_ERR("Failed to query red meter counter, meter id %u for eswitch%d",
                     doca_mtr_id, esw_id);
            return -1;
        }
        mtr_last_stats[doca_mtr_id].red = red_stats;

        green_counter_id = ovs_doca_get_post_meter_counter_id(doca_mtr_id,
                                                              DOCA_FLOW_METER_COLOR_GREEN);
        ret = query_counter(green_counter_id, &green_stats, error);
        if (ret) {
            VLOG_ERR("Failed to query green meter counter, meter id %u for eswitch%d",
                     doca_mtr_id, esw_id);
            return -1;
        }
        mtr_last_stats[doca_mtr_id].green = green_stats;
    }

    return 0;
}

int
ovs_doca_mtr_stats_read(uint16_t port_id OVS_UNUSED, uint32_t dp_meter_id,
                        struct rte_mtr_stats *stats,
                        struct rte_mtr_error *error)
{
    struct doca_flow_query *last_red_stats, *last_green_stats;
    struct doca_flow_query red_stats, green_stats;
    uint32_t red_counter_id, green_counter_id;
    uint64_t red_bytes, green_bytes;
    uint64_t red_pkts, green_pkts;
    uint32_t doca_mtr_id;
    uint32_t esw_id = 0; /* Only first eswitch is supported for now */
    int ret;

    /* id is determine by both the upper layer id, and the esw_id. */
    doca_mtr_id = ovs_doca_meter_id(dp_meter_id, esw_id);

    last_red_stats = &mtr_last_stats[doca_mtr_id].red;
    last_green_stats = &mtr_last_stats[doca_mtr_id].green;

    red_counter_id =
        ovs_doca_get_post_meter_counter_id(doca_mtr_id,
                                           DOCA_FLOW_METER_COLOR_RED);
    green_counter_id =
        ovs_doca_get_post_meter_counter_id(doca_mtr_id,
                                           DOCA_FLOW_METER_COLOR_GREEN);

    ret = query_counter(red_counter_id, &red_stats, error);
    if (ret) {
        VLOG_ERR("Failed to query red counter for meter id %u", doca_mtr_id);
        return -1;
    }

    ret = query_counter(green_counter_id, &green_stats, error);
    if (ret) {
        VLOG_ERR("Failed to query green counter for meter id %u", doca_mtr_id);
        return -1;
    }

    /* check for counter wraparound */
    red_pkts = ovs_u64_wrapsub(red_stats.total_pkts,
                               last_red_stats->total_pkts);
    red_bytes = ovs_u64_wrapsub(red_stats.total_bytes,
                                last_red_stats->total_bytes);
    green_pkts = ovs_u64_wrapsub(green_stats.total_pkts,
                                 last_green_stats->total_pkts);
    green_bytes = ovs_u64_wrapsub(green_stats.total_bytes,
                                  last_green_stats->total_bytes);

    stats->n_pkts[RTE_COLOR_GREEN] = green_pkts;
    stats->n_pkts[RTE_COLOR_YELLOW] = 0;
    stats->n_pkts[RTE_COLOR_RED] = red_pkts;

    stats->n_bytes[RTE_COLOR_GREEN] = green_bytes;
    stats->n_bytes[RTE_COLOR_YELLOW] = 0;
    stats->n_bytes[RTE_COLOR_RED] = red_bytes;

    stats->n_pkts_dropped = red_pkts;
    stats->n_bytes_dropped = red_bytes;

    return 0;
}

unsigned int
ovs_doca_max_ct_conns(void)
{
    return conntrack_offload_size();
}

int
ovs_doca_dev_probe(struct ovs_doca_netdev_data *dev_data,
                   const char *devargs,
                   bool already_valid)
{
    struct ds rte_devargs = DS_EMPTY_INITIALIZER;
    struct ovs_doca_esw_ctx_arg ctx_arg;
    struct ovs_doca_esw_key esw_key;
    struct ibv_pd *pd;
    int rv = 0;

    if (dev_data->esw_ctx) {
        /* Already probed. do nothing. */
        return 0;
    }

    if (ovs_doca_esw_key_parse(devargs, &esw_key)) {
        return EINVAL;
    }

    ctx_arg = (struct ovs_doca_esw_ctx_arg) {
        .esw_key = &esw_key,
        .dev_data = dev_data,
    };

    dev_data->esw_ctx = refmap_ref(ovs_doca_esw_rfm, &esw_key, &ctx_arg);
    if (!dev_data->esw_ctx) {
        VLOG_ERR("Could not get esw context for %s", devargs);
        return EINVAL;
    } else if (already_valid) {
        /* When a representor is probed before the eswitch, dpdk implicitly
         * probes the latter, thus probe is not called from
         * netdev_dpdk_process_devargs(). In this case we call probe at
         * ovs_doca_port_start(), to take the reference and initialize esw_ctx
         * handle in dev_data.
         */
        goto out;
    }

    if (doca_rdma_bridge_get_dev_pd(dev_data->esw_ctx->dev, &pd)) {
        VLOG_ERR("Could not get pd for %s", devargs);
        rv = EINVAL;
        goto out;
    }
    if (dev_data->esw_ctx->cmd_fd == -1) {
        dev_data->esw_ctx->cmd_fd = dup(pd->context->cmd_fd);
        if (dev_data->esw_ctx->cmd_fd == -1) {
            VLOG_ERR("Could not dup fd for %s. Error %s", devargs,
                     ovs_strerror(errno));
            rv = EBADF;
            goto out;
        }
    }

    ds_put_format(&rte_devargs, "%s,cmd_fd=%d,pd_handle=%u", devargs,
                  dev_data->esw_ctx->cmd_fd, pd->handle);

    VLOG_INFO("Probing '%s'", ds_cstr(&rte_devargs));
    if (rte_dev_probe(ds_cstr(&rte_devargs))) {
        rv = ENODEV;
        goto out;
    }

out:
    ds_destroy(&rte_devargs);
    if (rv) {
        ovs_doca_dev_close(dev_data);
    }
    return rv;
}

void
ovs_doca_dev_close(struct ovs_doca_netdev_data *dev_data)
{
    struct ovs_doca_esw_ctx *ctx = dev_data->esw_ctx;
    struct rte_eth_dev_info dev_info;
    char *pci_addr;
    bool last;
    int err;

    memset(&dev_info, 0, sizeof dev_info);

    if (rte_eth_dev_is_valid_port(dev_data->port_id)) {
        err = rte_eth_dev_info_get(dev_data->port_id, &dev_info);
        if (err) {
            VLOG_ERR("Failed to get info of port %d: %s", dev_data->port_id,
                     rte_strerror(-err));
        }
        err = rte_eth_dev_close(dev_data->port_id);
        if (err) {
            VLOG_ERR("Failed to close port %d: %s", dev_data->port_id,
                     rte_strerror(-err));
        }
    }

    pci_addr = xstrdup(ctx->pci_addr);
    last = refmap_unref(ovs_doca_esw_rfm, ctx);
    if (last && ctx->dev) {
        if (rte_eth_dev_is_valid_port(dev_data->esw_mgr_port_id)) {
            err = rte_eth_dev_close(dev_data->esw_mgr_port_id);
            if (err) {
                VLOG_ERR("Failed to close esw_mgr port %d: %s",
                         dev_data->esw_mgr_port_id, rte_strerror(-err));
            }
        }
        /* ctx->cmd_fd is closed inside. */
        if (dev_info.device) {
            err = rte_dev_remove(dev_info.device);
            if (err) {
                VLOG_ERR("Failed to remove device %s: %s", dev_data->devargs,
                         rte_strerror(-err));
            }
        }
        VLOG_INFO("Closing '%s'", pci_addr);
        err = doca_dev_close(ctx->dev);
        if (err) {
            VLOG_ERR("Failed to close doca dev %s. Error: %d (%s)", pci_addr,
                     err, doca_error_get_descr(err));
        }
        ctx->dev = NULL;
        ctx->cmd_fd = -1;
    }
    dev_data->esw_ctx = NULL;
    free(pci_addr);
}

static void
classify_in_port(struct dp_packet_batch *rx_batch,
                 struct dp_packet_batch *port_queue_batch[RTE_MAX_ETHPORTS],
                 uint16_t queue_id)
{
    struct dp_packet *pkt;
    uint32_t port_id;

    DP_PACKET_BATCH_FOR_EACH (i, pkt, rx_batch) {
        if (!dp_packet_has_flow_mark(pkt, &port_id)) {
            COVERAGE_INC(ovs_doca_no_mark);
            dp_packet_delete(pkt);
            continue;
        }
        dp_packet_reset_flow_mark(pkt);
        if (!rte_eth_dev_is_valid_port(port_id)) {
            COVERAGE_INC(ovs_doca_invalid_classify_port);
            dp_packet_delete(pkt);
            continue;
        }
        dp_packet_batch_add(&port_queue_batch[port_id][queue_id], pkt);
    }
}

int
ovs_doca_rx_burst(struct ovs_doca_netdev_data *dev_data,
                  uint16_t queue_id,
                  struct dp_packet_batch *batch)
{
    struct dp_packet_batch *port_batch;
    struct dp_packet_batch rx_batch;
    uint16_t esw_mgr_port_id;
    uint16_t port_id;

    esw_mgr_port_id = dev_data->esw_ctx->port_id;
    port_id = dev_data->port_id;

    if (port_id == esw_mgr_port_id) {
        rx_batch.count = rte_eth_rx_burst(esw_mgr_port_id, queue_id,
                                          (struct rte_mbuf **) rx_batch.packets,
                                          NETDEV_MAX_BURST);
        if (rx_batch.count == 0) {
            return 0;
        }
        classify_in_port(&rx_batch, dev_data->esw_ctx->port_queue_batch,
                         queue_id);
    }

    port_batch = &dev_data->esw_ctx->port_queue_batch[port_id][queue_id];
    memcpy(batch, port_batch, sizeof *batch);
    dp_packet_batch_init(port_batch);

    return batch->count;
}

void
ovs_doca_mgmt_queue_lock(void)
    OVS_ACQUIRES(mgmt_queue_lock)
{
    ovs_mutex_lock(&mgmt_queue_lock);
}

void
ovs_doca_mgmt_queue_unlock(void)
    OVS_RELEASES(mgmt_queue_lock)
{
    ovs_mutex_unlock(&mgmt_queue_lock);
}
